package ppJava.serie04.oceanlife;

import java.awt.BorderLayout;
import java.awt.Container;
import java.awt.GridLayout;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.MouseAdapter;
import java.awt.event.MouseEvent;
import java.io.File;

import java.util.LinkedList;
import java.util.concurrent.CyclicBarrier;

import javax.swing.JFileChooser;
import javax.swing.JFrame;
import javax.swing.JOptionPane;
import javax.swing.JPanel;
import javax.swing.JScrollPane;
import javax.swing.Timer;
import javax.swing.WindowConstants;

/**
 * OceanGUI is a JFrame GUI for the "oceanlife"-simulation.
 * 
 * It manages file-operations like load and save of ocean-data and all actions
 * which are performed. (MenuBar for file-ops and ToolBar for OceanObject
 * related actions)
 * 
 * @author ken,bpei
 * @see OceanInterface
 */
public class OceanGUI extends JFrame implements ActionListener {

	private static final long serialVersionUID = -913608740115972524L;
	// private static Logger logger =
	// Logger.getLogger("ppJava.serie04.oceanlife");

	/**
	 * FileChooser for load/save dialogs
	 */
	private JFileChooser fc = new JFileChooser();

	/**
	 * JPanel for drawing the ocean
	 */
	private OceanCanvasPanel canvasPanel;

	/**
	 * Toolbar for control of oceanObjects
	 */
	private OceanToolbar toolbar;

	private static int WIDTH = 1200;
	private static int HEIGHT = 800;

	private static int OCEAN_WIDTH = WIDTH;
	private static int OCEAN_DEPTH = HEIGHT;

	/**
	 * Determines if ToolBar is vertically aligned
	 */
	private static int TBVERTICAL = 0;

	/**
	 * This is the ocean which manages the oceanObjects
	 */
	private Ocean ocean;

	/**
	 * This object is generated if ToolBarItem is clicked and added if set a
	 * position in OceanCanvasPanel
	 */
	private OceanObject pendingObject;

	/**
	 * For Threading - Movement of objects is synchronized
	 */
	CyclicBarrier barrier;

	/**
	 * Timer for trigger repaint of JPanel and recalculate oceanObjects
	 */
	Timer timer;

	/**
	 * Pause between triggering of Timer
	 */
	int timing = 40;

	/**
	 * Initializes an ocean with fixed size and a list of OceanObjects
	 * 
	 * @param objList
	 * @see OceanObject
	 */
	private void initOcean(LinkedList<OceanObject> objList) {

		try {
			ocean = new Ocean(OCEAN_WIDTH, OCEAN_DEPTH, objList);
		} catch (OceanException e) {
			System.err.println("Invalid range.");
		}
	}

	/**
	 * Removes all oceanObjects from GUI-related interfaces and clears the list
	 * of oceanObjects
	 */
	private void clearOcean() {
		toolbar.getToolComboBox().removeAllItems();
		ocean.clearObjectList();
	}

	/**
	 * Removes a specific oceanObject at an given index (Removes this also from
	 * GUI)
	 * 
	 * @param index
	 * @throws OceanException
	 */
	private void removeObject(int index) throws OceanException {
		System.out.println("remove: "
				+ ocean.getOceanObjects().get(index).getClassName()
				+ " at index: " + index + " size: "
				+ ocean.getOceanObjects().size());

		ocean.removeOceanObject(index);
		toolbar.getToolComboBox().removeItemAt(index);
	}

	/**
	 * Manages action performed on a toolBar (create/remove-Objects and control
	 * of movement)
	 */
	private ActionListener toolBarActionListener = new ActionListener() {

		@Override
		public void actionPerformed(ActionEvent event) {

			try {

				if ("createOceanRock".equals(event.getActionCommand())) {
					pendingObject = new Stone(0, 0);
					((Stone) pendingObject).setMoveLimit(HEIGHT);
				}

				if ("createOceanPlant".equals(event.getActionCommand())) {
					pendingObject = new Plant(0, 0);
					((Plant) pendingObject).setMoveLimit(HEIGHT);
				}

				if ("createOceanFish".equals(event.getActionCommand())) {
					pendingObject = new Fish(0, 0);
				}

				if ("createOceanBubble".equals(event.getActionCommand())) {
					pendingObject = new Bubble(0, 0);
					((Bubble) pendingObject).setMoveLimit(HEIGHT);
				}

				if ("removeOceanObject".equals(event.getActionCommand())
						&& toolbar.getToolComboBox().getSelectedIndex() >= 0) {
					removeObject(toolbar.getToolComboBox().getSelectedIndex());
				}

				if ("stopMovement".equals(event.getActionCommand())) {
					timer.stop();
					System.out.println("stopMovement");
				}

				if ("startMovement".equals(event.getActionCommand())) {
					timer.start();
					System.out.println("startMovement");
				}

				if ("traceMovement".equals(event.getActionCommand())) {

					timer.start();
					startOceanThreads();
					canvasPanel.repaint();
					try {
						Thread.sleep(100);
						timer.stop();
					} catch (InterruptedException e) {
						JOptionPane.showMessageDialog(canvasPanel, e
								.getMessage());
					}

					System.out.println("traceMovement");
				}

			} catch (OceanException e) {
				JOptionPane.showMessageDialog(canvasPanel, e.getMessage());
			}

			// refresh
			toolbar.validate();
		}
	};

	/**
	 * Manages action performed on the menuBar (load/save-file and quit)
	 */
	private ActionListener menuBarActionListener = new ActionListener() {

		@Override
		public void actionPerformed(ActionEvent event) {
			try {

				if ("newOceanFile".equals(event.getActionCommand())) {
					clearOcean();
				}
				if ("loadOceanFile".equals(event.getActionCommand())) {

					int returnVal = fc.showOpenDialog(OceanGUI.this);

					if (returnVal == JFileChooser.APPROVE_OPTION) {

						clearOcean();

						File file = fc.getSelectedFile();
						System.out.println("Opening: " + " file:/"
								+ file.getAbsolutePath() + "\n");
						ocean = (Ocean) ocean.loadOcean(file.getAbsolutePath());
						refreshComboList();

					} else {
						System.out.println("Open command cancelled by user.\n");
					}

					System.out.println("loadOceanFile");
				}
				if ("saveOceanFile".equals(event.getActionCommand())) {

					int returnVal = fc.showSaveDialog(OceanGUI.this);

					if (returnVal == JFileChooser.APPROVE_OPTION) {
						File file = fc.getSelectedFile();
						System.out.println("Saving: file:/"
								+ file.getAbsolutePath() + ".\n");
						ocean.saveOcean(file.getAbsolutePath());
					} else {
						System.out.println("Save command cancelled by user.\n");
					}
					System.out.println("saveOceanFile");
				}
				if ("quitGUI".equals(event.getActionCommand())) {
					System.exit(0);
				}
				canvasPanel.repaint();

			} catch (Exception e) {
				e.printStackTrace();
			}
		}
	};

	/**
	 * Adds a oceanObject to ComboBox (of toolBar) and appends it to ocean
	 * 
	 * @param oceanObject
	 * @throws OceanException
	 *             (i.e. invalid position)
	 */
	public void addNewObject(OceanObject oceanObject) throws OceanException {
		ocean.addOceanObject(oceanObject);
		toolbar.getToolComboBox().addItem(oceanObject);
		toolbar.getToolComboBox().setSelectedItem(oceanObject);
		System.out.println("create: " + oceanObject.getClassName());
	}

	/**
	 * Refresh records of ComboList
	 */
	public void refreshComboList() {
		for (OceanObject oceanObject : ocean.getOceanObjects()) {
			toolbar.getToolComboBox().addItem(oceanObject);
		}
	}

	/**
	 * Manages the "mouseClicked"-event. If there is a pendingObject, it will be
	 * added to the ocean otherwise action is ignored. (If it fails a message is
	 * displayed and object will be ignored)
	 * 
	 */
	public class OceanMouseListener extends MouseAdapter {

		@Override
		public void mouseClicked(MouseEvent mouseEvent) {
			if (pendingObject != null) {
				try {
					int position[] = { mouseEvent.getX(), mouseEvent.getY() };

					if (pendingObject instanceof Plant) {
						position[1] = HEIGHT;
					}

					pendingObject.setPosition(position);
					addNewObject(pendingObject);
					pendingObject = null;

					canvasPanel.repaint();

				} catch (OceanException e) {
					JOptionPane.showMessageDialog(canvasPanel, e.getMessage());
				}
			}
		}
	}

	/**
	 * Builds all components of GUI. (MenuBar, Toolbar, Panel for drawing)
	 * 
	 * All items of these components are defined at their own classes. i.e.
	 * icons for creation of objects are defined in ToolBar-class. All
	 * ActionListeners are defined in this class and passed
	 * 
	 * @see OceanMenuBar
	 * @see OceanToolbar
	 * @see OceanCanvasPanel
	 * 
	 * @see menuBarActionListener
	 * @see toolBarActionListener
	 * @see OceanMouseListener
	 */
	private void buildGUI() {
		initOcean(new LinkedList<OceanObject>());

		OceanMenuBar menuBar = new OceanMenuBar();
		menuBar.setActionListener(menuBarActionListener);
		menuBar.populateMenu();
		this.setJMenuBar(menuBar);

		/* this takes all items */
		Container content = getContentPane();
		content.setLayout(new BorderLayout());
		this.setSize(WIDTH, HEIGHT);

		/* a draggable toolbar for (create / remove ... */
		toolbar = new OceanToolbar();
		toolbar.setActionListener(toolBarActionListener);
		toolbar.populateToolbar();
		toolbar.setOrientation(TBVERTICAL);
		content.add(toolbar, BorderLayout.SOUTH);

		canvasPanel = new OceanCanvasPanel();
		canvasPanel.setSize(OCEAN_WIDTH, OCEAN_DEPTH);
		canvasPanel.setOcean(ocean);
		canvasPanel.addMouseListener(new OceanMouseListener());

		/* put items into sth like a table */
		JPanel contentDisplay = new JPanel(new GridLayout(0, 1));
		contentDisplay.add(canvasPanel);
		// contentDisplay.add(txtOut);
		contentDisplay.setSize(WIDTH, HEIGHT);

		// needed for the _dragging_ of toolbar
		JScrollPane scrollPane = new JScrollPane(contentDisplay);
		content.add(scrollPane, BorderLayout.CENTER);

		// Closing the Window by clicking on [x]
		setDefaultCloseOperation(WindowConstants.EXIT_ON_CLOSE);

	}

	/**
	 * Triggers recalculating and repaint
	 */
	private void startTimer() {
		timer = new Timer(timing, this);
		timer.start();
	}

	private void startOceanThreads() {
		for (OceanObject oceanObject : ocean.getOceanObjects()) {
			oceanObject.setBarrier(barrier);
			new Thread(oceanObject).start();
		}
	}

	/**
	 * Repaints Panel and recalculates collisions
	 */
	@Override
	public void actionPerformed(ActionEvent arg0) {
		// repaint the ocean
		canvasPanel.repaint();
		// checkObjects - handle collisions
		this.checkObjectPositions(ocean.getOceanObjects());

		int N = ocean.getOceanObjects().size();

		if (N > 0) {
			barrier = new CyclicBarrier(N);
			startOceanThreads();
		}
	}

	/**
	 * Collision detection (i.e. changing movement of oceanObjects)
	 * 
	 * @param oceanObjects
	 *            - list of objects which are checked
	 */
	public void checkObjectPositions(LinkedList<OceanObject> oceanObjects) {

		if (oceanObjects.size() > 1) {
			for (int i = 0; i < oceanObjects.size(); i++) {
				OceanObject tempObject = oceanObjects.get(i);
				int nextObjectId = 0;

				// only fishes are able to move or "react"
				if (tempObject instanceof Fish) {

					try {

						nextObjectId = ocean.indexOfNearestObject(tempObject
								.getPosition());
						OceanObject tempRefObject = oceanObjects
								.get(nextObjectId);

						// collides (with tempRefObject)
						if (tempObject.getDistance(tempRefObject.getPosition()) <= (tempObject
								.getOffset()[1] + tempRefObject.getOffset()[1])) {

							// what happens if a fish meets bubbles?
							if (tempRefObject instanceof Bubble) {

							} else {

								// a fish meets another
								if ((tempRefObject instanceof Fish)
										&& (tempObject
												.isVulnerableTo(tempRefObject))) {

									// tempRefObj.doEatAnim
									removeObject(i);

								} else {
									// a fish meets a plant or rock
									tempObject.changeDirections(tempRefObject);
								}
							}
						}

					} catch (OceanException e) {
						// ocean is empty
					}
				}
			}
		}
	}

	/**
	 * Called first - builds the GUI and starts a Timer which triggers control
	 * and drawing of the ocean
	 * 
	 * @param args
	 *            -- ignored
	 */
	public static void main(String[] args) {

		OceanGUI oceanGUI = new OceanGUI();
		oceanGUI.buildGUI();
		oceanGUI.setVisible(true);
		oceanGUI.startTimer();

	}

}
