package ppJava.serie04.oceanlife;

import java.awt.Image;
import java.io.Serializable;
import java.util.Arrays;
import java.util.Random;
import java.util.concurrent.BrokenBarrierException;
import java.util.concurrent.CyclicBarrier;

import javax.swing.ImageIcon;

/**
 * Objects in the Ocean
 * 
 * @author bpei,ken
 * 
 */
public abstract class OceanObject implements Runnable, Serializable {

	private static final long serialVersionUID = 1L;

	private int[] position = new int[2];
	private int[] moveVector = new int[2];
	private int[] borders = new int[2];
	private int[] offset = new int[2];

	private static int objectCount = 0;

	private transient CyclicBarrier barrier;

	private String imagePath = null;
	private ImageIcon imageIcon = null;
	private int imageSize = 50 + new Random().nextInt(100);

	public String getClassName() {
		return this.getClass().getName();
	}

	/**
	 * Generates a random vector for moving
	 * 
	 * @return a random vector for moving
	 */
	public int[] genRandomVector() {
		int moveVector[] = { 0, 0 };
		moveVector[0] = ((int) (Math.random() * 3) + 1)
				* (((Math.random() - 0.5) < 0) ? -1 : 1);
		moveVector[1] = ((int) (Math.random()) + 1)
				* (((Math.random() - 0.5) < 0) ? -1 : 1);
		return moveVector;
	}

	public void checkBarrier() {
		try {
			getBarrier().await();
		} catch (InterruptedException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		} catch (BrokenBarrierException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}
	}

	/**
	 * Constructor for an Object, increases the object Count
	 */
	public OceanObject() {
		objectCount++;
	}

	/**
	 * Get the object count
	 * 
	 * @return the object count
	 */
	public static int getObjectCount() {
		return objectCount;
	}

	/**
	 * get the position of an object
	 * 
	 * @return the position of an object
	 */
	public int[] getPosition() {
		return position;
	}

	/**
	 * Sets the position of an object
	 * 
	 * @param position
	 */
	public void setPosition(int[] position) {
		this.position = position;
	}

	/**
	 * Increase the X-Value of the position by i
	 * 
	 * @param i
	 */
	public void incPosX(int i) {
		this.position[0] += i;
	}

	/**
	 * Increase the Y-Value of the position by i
	 * 
	 * @param i
	 */
	public void incPosY(int i) {
		this.position[1] += i;
	}

	/**
	 * Switch the direction in which to move
	 * 
	 * @param signs
	 *            of the movement before switching
	 */
	public void switchMovementDirection(int signs[]) {
		int moveVector[] = this.getMoveVector();

		moveVector[0] *= (signs[0] * this.moveVector[0] > 0) ? 1 : -1;
		moveVector[1] *= (signs[1] * this.moveVector[1] > 0) ? 1 : -1;

		this.setMoveVector(moveVector);
	}

	/**
	 * Get the Distance to a position
	 * 
	 * @param position
	 * @return the distance to a position
	 */
	public int getDistance(int position[]) {
		return calcAbsDistance(this.getPosition(), position);
	}

	/**
	 * Calculate the distance between two positions
	 * 
	 * @param pos1
	 * @param pos2
	 * @return the distance between two positions
	 */
	public int calcAbsDistance(int pos1[], int pos2[]) {
		return (int) Math.sqrt((Math.pow(Math.abs(pos1[0] - pos2[0]), 2) + Math
				.pow(Math.abs(pos1[1] - pos2[1]), 2)));
	}

	/**
	 * Calculate the difference of the X- and Y-Values of two positions
	 * 
	 * @param position
	 * @return the difference of the X- and Y-Values of two positions
	 */
	public int[] getDiffVector(int position[]) {
		int diffPos[] = { this.getPosition()[0] - position[0],
				this.getPosition()[1] - position[1] };
		return diffPos;
	}

	/**
	 * Check if the position is in the Ocean
	 * 
	 * @param ocean
	 * @throws OceanException
	 *             - if the position is not in the Ocean
	 */
	public void checkPosition(Ocean ocean) throws OceanException {
		int pos[] = this.getPosition();
		if ((pos[0] > ocean.getWidth() || pos[1] > ocean.getDepth())
				|| (pos[0] < 0 || pos[1] < 0))
			throw new OceanException("Invalid position.");
	}

	/**
	 * Make one step in movement
	 */
	public void moveStep() {
		this.incPosX(moveVector[0]);
		this.incPosY(moveVector[1]);
	}

	/**
	 * Change the direction in which to move of the object
	 * 
	 * @param tempRefObject
	 */
	public void changeDirections(OceanObject tempRefObject) {

		int diffVector[] = this.getDiffVector(tempRefObject.getPosition());
		int moveVector[] = diffVector;
		this.switchMovementDirection(moveVector);
	}

	/**
	 * Have two objects the same X-direction?
	 * 
	 * @param tempRefObject
	 * @return true if they move in the same X-direction
	 */
	public boolean hasSameDirection(OceanObject tempRefObject) {
		return ((this.moveVector[0] * tempRefObject.getMoveVector()[0]) > 0);
	}

	/**
	 * Can an object be eaten by another object?
	 * 
	 * @param tempRefObject
	 * @return true if it can be eaten
	 */
	public boolean isVulnerableTo(OceanObject tempRefObject) {
		return (this.hasSameDirection(tempRefObject) && ((this.moveVector[0] > 0 && (this.position[0] > tempRefObject
				.getPosition()[0])) || (this.moveVector[0] < 0 && (this.position[0] < tempRefObject
				.getPosition()[0]))));
	}

	/**
	 * Returns an ImageIcon, or null if the path was invalid. (source: oracle)
	 */
	protected ImageIcon createImageIcon(String path, String description) {
		java.net.URL imgURL = getClass().getResource(path);
		if (imgURL != null) {
			return new ImageIcon(imgURL, description);
		} else {
			System.err.println("Couldn't find file: " + path);
			return null;
		}
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see java.lang.Object#toString()
	 */
	@Override
	public String toString() {
		return "[position:" + Arrays.toString(position) + "]";
	}

	/**
	 * Set the path of the Image to the Object
	 * 
	 * @param imageResource
	 */
	public void setImageResource(String imageResource) {
		this.imagePath = imageResource;
	}

	/**
	 * Return the path of the Image to the Object
	 * 
	 * @return the path of the Image to the Object
	 */
	public String getImageResource() {
		return imagePath;
	}

	/**
	 * Sets the move vector of the object
	 * 
	 * @param moveVector
	 */
	public void setMoveVector(int[] moveVector) {
		this.moveVector = moveVector;
	}

	/**
	 * Return the move vector of the object
	 * 
	 * @return the move vector of the object
	 */
	public int[] getMoveVector() {
		return moveVector;
	}

	/**
	 * Sets the borders in which the object is allowed to move
	 * 
	 * @param borders
	 */
	public void setBorders(int[] borders) {
		this.borders = borders;
	}

	/**
	 * Return the borders in which the object is allowed to move
	 * 
	 * @return the borders in which the object is allowed to move
	 */
	public int[] getBorders() {
		return borders;
	}

	/**
	 * Sets the width and the height of the object
	 * 
	 * @param offset
	 */
	public void setOffset(int[] offset) {
		this.offset = offset;
	}

	/**
	 * Return the width and the height of the object
	 * 
	 * @return the width and the height of the object
	 */
	public int[] getOffset() {
		return offset;
	}

	/**
	 * Resizes the sprite of oceanObject
	 */
	public void normalizeImage() {
		int ratio[] = { imageIcon.getIconWidth(), imageIcon.getIconHeight() };

		Image newimg = imageIcon.getImage().getScaledInstance(
				imageSize * ratio[0] / ratio[1], imageSize,
				java.awt.Image.SCALE_SMOOTH);
		this.imageIcon = new ImageIcon(newimg);
	}

	/**
	 * Sets the Image to the object
	 * 
	 * @param image
	 */
	public void setImageIcon(ImageIcon image) {
		this.imageIcon = image;
		this.normalizeImage();
	}

	/**
	 * Return the Image to the object
	 * 
	 * @return the Image to the object
	 */
	public ImageIcon getImageIcon() {
		return imageIcon;
	}

	/**
	 * Return the size of the Image to the Object
	 * 
	 * @return the size of the Image to the Object
	 */
	public int getImageSize() {
		return imageSize;
	}

	/**
	 * Sets the size of the Image
	 * 
	 * @param imageSize
	 */
	public void setImageSize(int imageSize) {
		this.imageSize = imageSize;
	}

	/**
	 * @param barrier
	 *            the barrier to set
	 */
	public void setBarrier(CyclicBarrier barrier) {
		this.barrier = barrier;
	}

	/**
	 * @return the barrier
	 */
	public CyclicBarrier getBarrier() {
		return barrier;
	}

}